@@ -60,9 +60,14 @@ typedef struct {
     GtkWidget *stack;
     GtkWidget *swin_view;
     GtkWidget *swin_preview;
+    GtkWidget *swin_info;
     GtkWidget *icon_view;
+    GtkWidget *search_bar;
+    GtkWidget *search_entry;
+    GtkWidget *search_toggle;
 
     GtkTreeModel *model;
+    GtkTreeModel *filter_model;
 
     GFile *font_file;
 } FontViewApplication;
@@ -163,6 +168,8 @@ add_row (GtkWidget *grid,
          gboolean multiline)
 {
     GtkWidget *name_w, *label;
+    int i;
+    const char *p;
 
     name_w = gtk_label_new (name);
     gtk_style_context_add_class (gtk_widget_get_style_context (name_w), "dim-label");
@@ -181,12 +188,24 @@ add_row (GtkWidget *grid,
 
     gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_END);
     gtk_label_set_max_width_chars (GTK_LABEL (label), 64);
-    if (multiline && g_utf8_strlen (value, -1) > 64) {
-       gtk_label_set_width_chars (GTK_LABEL (label), 64);
-    }
-    gtk_label_set_lines (GTK_LABEL (label), multiline ? 10 : 1);
 
-    gtk_grid_attach_next_to (GTK_GRID (grid), label, 
+    if (multiline && g_utf8_strlen (value, -1) > 64) {
+        gtk_label_set_width_chars (GTK_LABEL (label), 64);
+        gtk_label_set_lines (GTK_LABEL (label), 10);
+
+        p = value;
+        i = 0;
+        while (p) {
+            p = strchr (p + 1, '\n');
+            i++;
+        }
+        if (i > 3) { /* multi-paragraph text */
+            gtk_label_set_ellipsize (GTK_LABEL (label), PANGO_ELLIPSIZE_NONE);
+            gtk_label_set_lines (GTK_LABEL (label), -1);
+        }
+    }
+
+    gtk_grid_attach_next_to (GTK_GRID (grid), label,
                              name_w, GTK_POS_RIGHT,
                              1, 1);
 }
@@ -685,26 +704,21 @@ info_button_clicked_cb (GtkButton *button,
                         gpointer user_data)
 {
     FontViewApplication *self = user_data;
-    GtkWidget *stack, *switcher, *grid, *dialog;
+    GtkWidget *grid;
+    GtkWidget *child;
     FT_Face face = sushi_font_widget_get_ft_face (SUSHI_FONT_WIDGET (self->font_widget));
 
+    if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (button))) {
+        gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "preview");
+        return;
+    }
+
     if (face == NULL)
         return;
 
-    dialog = g_object_new (GTK_TYPE_DIALOG,
-                           "transient-for", self->main_window,
-                           "modal", TRUE,
-                           "resizable", FALSE,
-                           "destroy-with-parent", TRUE,
-                           "use-header-bar", TRUE,
-                           NULL);
-    stack = gtk_stack_new ();
-    gtk_container_add (GTK_CONTAINER (gtk_dialog_get_content_area (GTK_DIALOG (dialog))), stack);
-
-    switcher = gtk_stack_switcher_new ();
-    gtk_stack_switcher_set_stack (GTK_STACK_SWITCHER (switcher), GTK_STACK (stack));
-    gtk_header_bar_set_custom_title (GTK_HEADER_BAR (gtk_window_get_titlebar (GTK_WINDOW (dialog))),
-                                     switcher);
+    child = gtk_bin_get_child (GTK_BIN (self->swin_info));
+    if (child)
+        gtk_widget_destroy (child);
 
     grid = gtk_grid_new ();
     gtk_orientable_set_orientation (GTK_ORIENTABLE (grid), GTK_ORIENTATION_VERTICAL);
@@ -713,22 +727,11 @@ info_button_clicked_cb (GtkButton *button,
     gtk_grid_set_row_spacing (GTK_GRID (grid), 2);
 
     populate_grid (self, grid, face);
-
-    gtk_stack_add_titled (GTK_STACK (stack), grid, "info", _("General"));
-
-    grid = gtk_grid_new ();
-    gtk_orientable_set_orientation (GTK_ORIENTABLE (grid), GTK_ORIENTATION_VERTICAL);
-    g_object_set (grid, "margin", 20, NULL);
-    gtk_grid_set_column_spacing (GTK_GRID (grid), 8);
-    gtk_grid_set_row_spacing (GTK_GRID (grid), 2);
-
     populate_details (self, grid, face);
+    gtk_container_add (GTK_CONTAINER (self->swin_info), grid);
 
-    gtk_stack_add_titled (GTK_STACK (stack), grid, "details", _("Details"));
-
-    g_signal_connect (dialog, "response",
-                      G_CALLBACK (gtk_widget_destroy), NULL);
-    gtk_widget_show_all (dialog);
+    gtk_widget_show_all (self->swin_info);
+    gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "info");
 }
 
 static void
@@ -739,6 +742,39 @@ font_view_update_scale_factor (FontViewApplication *self)
                                       scale_factor);
 }
 
+static gboolean
+font_visible_func (GtkTreeModel *model,
+                   GtkTreeIter  *iter,
+                   gpointer      data)
+{
+  FontViewApplication *self = data;
+  gboolean ret;
+  const char *search;
+  char *name;
+  char *cf_name;
+  char *cf_search;
+
+  if (!gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (self->search_toggle)))
+    return TRUE;
+
+  search = gtk_entry_get_text (GTK_ENTRY (self->search_entry));
+
+  gtk_tree_model_get (model, iter,
+                      COLUMN_NAME, &name,
+                      -1);
+
+  cf_name = g_utf8_casefold (name, -1);
+  cf_search = g_utf8_casefold (search, -1);
+
+  ret = strstr (cf_name, cf_search) != NULL;
+
+  g_free (name);
+  g_free (cf_name);
+  g_free (cf_search);
+
+  return ret;
+}
+
 static void
 font_view_ensure_model (FontViewApplication *self)
 {
@@ -746,6 +782,9 @@ font_view_ensure_model (FontViewApplication *self)
         return;
 
     self->model = font_view_model_new ();
+    self->filter_model = gtk_tree_model_filter_new (self->model, NULL);
+    gtk_tree_model_filter_set_visible_func (GTK_TREE_MODEL_FILTER (self->filter_model),
+                                            font_visible_func, self, NULL);
     g_signal_connect (self->model, "config-changed",
                       G_CALLBACK (font_model_config_changed_cb), self);
 
@@ -777,13 +816,13 @@ font_view_application_do_open (FontViewApplication *self,
     }
 
     if (self->info_button == NULL) {
-        self->info_button = gtk_button_new_with_label (_("Info"));
+        self->info_button = gtk_toggle_button_new_with_label (_("Info"));
         gtk_widget_set_valign (self->info_button, GTK_ALIGN_CENTER);
         gtk_style_context_add_class (gtk_widget_get_style_context (self->info_button),
                                      "text-button");
         gtk_header_bar_pack_end (GTK_HEADER_BAR (self->header), self->info_button);
 
-        g_signal_connect (self->info_button, "clicked",
+        g_signal_connect (self->info_button, "toggled",
                           G_CALLBACK (info_button_clicked_cb), self);
     }
 
@@ -802,11 +841,18 @@ font_view_application_do_open (FontViewApplication *self,
                           G_CALLBACK (back_button_clicked_cb), self);
     }
 
+    gtk_widget_hide (self->search_toggle);
+
     uri = g_file_get_uri (file);
 
     if (self->font_widget == NULL) {
+        GtkWidget *viewport;
+
         self->font_widget = GTK_WIDGET (sushi_font_widget_new (uri, face_index));
         gtk_container_add (GTK_CONTAINER (self->swin_preview), self->font_widget);
+        viewport = gtk_widget_get_parent (self->font_widget);
+        gtk_scrollable_set_hscroll_policy (GTK_SCROLLABLE (viewport), GTK_SCROLL_NATURAL);
+        gtk_scrollable_set_vscroll_policy (GTK_SCROLLABLE (viewport), GTK_SCROLL_NATURAL);
 
         g_signal_connect (self->font_widget, "loaded",
                           G_CALLBACK (font_widget_loaded_cb), self);
@@ -821,6 +867,7 @@ font_view_application_do_open (FontViewApplication *self,
 
     gtk_widget_show_all (self->main_window);
     gtk_stack_set_visible_child_name (GTK_STACK (self->stack), "preview");
+    gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (self->info_button), FALSE);
 }
 
 static gboolean
@@ -830,6 +877,7 @@ icon_view_release_cb (GtkWidget *widget,
 {
     FontViewApplication *self = user_data;
     GtkTreePath *path;
+    GtkTreeIter filter_iter;
     GtkTreeIter iter;
     gchar *font_path;
     gint face_index;
@@ -843,7 +891,10 @@ icon_view_release_cb (GtkWidget *widget,
                                           event->x, event->y);
 
     if (path != NULL &&
-        gtk_tree_model_get_iter (self->model, &iter, path)) {
+        gtk_tree_model_get_iter (self->filter_model, &filter_iter, path)) {
+        gtk_tree_model_filter_convert_iter_to_child_iter (GTK_TREE_MODEL_FILTER (self->filter_model),
+                                                          &iter,
+                                                          &filter_iter);
         gtk_tree_model_get (self->model, &iter,
                             COLUMN_PATH, &font_path,
                             COLUMN_FACE_INDEX, &face_index,
@@ -881,6 +932,8 @@ font_view_application_do_overview (FontViewApplication *self)
         self->install_button = NULL;
     }
 
+    gtk_widget_show (self->search_toggle);
+
     font_view_ensure_model (self);
 
     gtk_header_bar_set_title (GTK_HEADER_BAR (self->header), _("All Fonts"));
@@ -890,7 +943,7 @@ font_view_application_do_overview (FontViewApplication *self)
         GtkWidget *icon_view;
         GtkCellRenderer *cell;
 
-        self->icon_view = icon_view = gtk_icon_view_new_with_model (self->model);
+        self->icon_view = icon_view = gtk_icon_view_new_with_model (self->filter_model);
 
         g_object_set (icon_view,
                       "column-spacing", VIEW_COLUMN_SPACING,
@@ -943,7 +996,7 @@ font_view_window_key_press_event_cb (GtkWidget *widget,
         return TRUE;
     }
 
-    return FALSE;
+    return gtk_search_bar_handle_event (GTK_SEARCH_BAR (self->search_bar), (GdkEvent *)event);
 }
 
 static void
@@ -1024,10 +1077,17 @@ static GActionEntry action_entries[] = {
     { "quit", action_quit, NULL, NULL, NULL }
 };
 
+static void
+search_text_changed (GtkEntry *entry,
+                     FontViewApplication *self)
+{
+  gtk_tree_model_filter_refilter (GTK_TREE_MODEL_FILTER (self->filter_model));
+}
+
 static void
 ensure_window (FontViewApplication *self)
 {
-    GtkWidget *window, *swin;
+    GtkWidget *window, *swin, *box, *image;
 
     if (self->main_window)
         return;
@@ -1055,16 +1115,42 @@ ensure_window (FontViewApplication *self)
     gtk_widget_set_hexpand (self->stack, TRUE);
     gtk_widget_set_vexpand (self->stack, TRUE);
 
+    box = gtk_box_new (GTK_ORIENTATION_VERTICAL, 0);
+    gtk_stack_add_named (GTK_STACK (self->stack), box, "overview");
+
+    self->search_bar = gtk_search_bar_new ();
+    gtk_container_add (GTK_CONTAINER (box), self->search_bar);
+    self->search_entry = gtk_search_entry_new ();
+    gtk_entry_set_width_chars (GTK_ENTRY (self->search_entry), 40);
+    gtk_container_add (GTK_CONTAINER (self->search_bar), self->search_entry);
+    self->search_toggle = gtk_toggle_button_new ();
+    gtk_widget_set_no_show_all (self->search_toggle, TRUE);
+    gtk_widget_show (self->search_toggle);
+    image = gtk_image_new_from_icon_name ("edit-find-symbolic", GTK_ICON_SIZE_BUTTON);
+    gtk_widget_show (image);
+    gtk_container_add (GTK_CONTAINER (self->search_toggle), image);
+    gtk_header_bar_pack_end (GTK_HEADER_BAR (self->header), self->search_toggle);
+    g_object_bind_property (self->search_bar, "search-mode-enabled",
+                            self->search_toggle, "active",
+                            G_BINDING_BIDIRECTIONAL);
+
+    g_signal_connect (self->search_entry, "search-changed", G_CALLBACK (search_text_changed), self);
+
     self->swin_view = swin = gtk_scrolled_window_new (NULL, NULL);
     gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swin),
 				    GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
-    gtk_stack_add_named (GTK_STACK (self->stack), swin, "overview");
+    gtk_container_add (GTK_CONTAINER (box), self->swin_view);
 
     self->swin_preview = swin = gtk_scrolled_window_new (NULL, NULL);
     gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swin),
-         			    GTK_POLICY_AUTOMATIC, GTK_POLICY_NEVER);
+         			    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
     gtk_stack_add_named (GTK_STACK (self->stack), swin, "preview");
 
+    self->swin_info = swin = gtk_scrolled_window_new (NULL, NULL);
+    gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (swin),
+         			    GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
+    gtk_stack_add_named (GTK_STACK (self->stack), swin, "info");
+
     gtk_widget_show_all (window);
 }
 
@@ -1106,6 +1192,7 @@ font_view_application_dispose (GObject *obj)
     FontViewApplication *self = FONT_VIEW_APPLICATION (obj);
 
     g_clear_object (&self->font_file);
+    g_clear_object (&self->filter_model);
     g_clear_object (&self->model);
 
     G_OBJECT_CLASS (font_view_application_parent_class)->dispose (obj);
